/*
  Copyright 2003 Savonet team

  This file is part of Ocaml-shout.
  
  Ocaml-shout is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
   
  Ocaml-shout is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with Ocaml-shout; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
 * Libshout 2 bindings for OCaml.
 *
 * @author Samuel Mimram
 *
 */

/* $Id: shout_stubs.c,v 1.7 2004/06/01 12:42:13 smimram Exp $ */

#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/misc.h>
#include <caml/mlvalues.h>

#include <stdio.h>
#include <shout/shout.h>


static shout_t* shout_of_block(value block)
{
  return (shout_t*)Field(block, 1);
}

static void finalize_shout_t (value block)
{
  shout_t *s = shout_of_block(block);
  shout_free(s);
}

CAMLprim value ocaml_shout_init()
{
  CAMLparam0();
  shout_init();
  CAMLreturn(Val_unit);
}

CAMLprim value ocaml_shout_shutdown()
{
  CAMLparam0();
  shout_shutdown();
  CAMLreturn(Val_unit);
}

CAMLprim value ocaml_shout_version()
{
  CAMLparam0();
  CAMLlocal1(block);
  int major, minor, patch;
  const char *version = shout_version(&major, &minor, &patch);
  block = alloc_tuple(4);
  Field(block, 0) = copy_string(version);
  Field(block, 1) = Val_int(major);
  Field(block, 2) = Val_int(minor);
  Field(block, 3) = Val_int(patch);
  CAMLreturn(block);
}

CAMLprim value ocaml_shout_new ()
{
  CAMLparam0();
  CAMLlocal1(block);
  shout_t *s = shout_new();
  /*  150: beurk beurk beurk */
  block = alloc_final(2, finalize_shout_t, 150, 1000);
  if (s == NULL)
    raise_constant(*caml_named_value("shout_exn_malloc"));
  Field(block, 1) = (value)s;
  CAMLreturn(block);
}

CAMLprim value ocaml_shout_get_error(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *error = shout_get_error(s);
  CAMLreturn(copy_string(error));
}

CAMLprim value ocaml_shout_get_errno(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int error = shout_get_errno(s);
  CAMLreturn(Val_int(error));
}

CAMLprim value ocaml_shout_get_connected(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  CAMLreturn(shout_get_connected(s)==SHOUTERR_CONNECTED?Val_true:Val_false);
}

static void check_errors(int err)
{
  switch (err)
    {
    case SHOUTERR_SUCCESS:
      return;
      
    case SHOUTERR_INSANE:
      raise_constant(*caml_named_value("shout_exn_insane"));
      break;
      
    case SHOUTERR_NOCONNECT:
      raise_constant(*caml_named_value("shout_exn_no_connect"));
      break;
      
    case SHOUTERR_NOLOGIN:
      raise_constant(*caml_named_value("shout_exn_no_login"));
      break;
	
    case SHOUTERR_SOCKET:
      raise_constant(*caml_named_value("shout_exn_socket"));
      break;
      
    case SHOUTERR_MALLOC:
      raise_constant(*caml_named_value("shout_exn_malloc"));
      break;
      
    case SHOUTERR_METADATA:
      raise_constant(*caml_named_value("shout_exn_metadata"));
      break;
      
    case SHOUTERR_CONNECTED:
      raise_constant(*caml_named_value("shout_exn_connected"));
      break;
	
    case SHOUTERR_UNCONNECTED:
      raise_constant(*caml_named_value("shout_exn_unconnected"));
      break;
      
    case SHOUTERR_UNSUPPORTED:
      raise_constant(*caml_named_value("shout_exn_unsupported"));
      break;

    default:
      break;
    }
}

static value unit_or_error(int err)
{
  check_errors(err);
  return Val_unit;
}

CAMLprim value ocaml_shout_set_host(value block, value host)
{
  CAMLparam2(block, host);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_host(s, String_val(host));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_host(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *host = shout_get_host(s);
  CAMLreturn(copy_string(host));
}

CAMLprim value ocaml_shout_set_port(value block, value port)
{
  CAMLparam2(block, port);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_port(s, (unsigned short)Int_val(port));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_port(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int port = shout_get_port(s);
  CAMLreturn(Val_int(port));
}

CAMLprim value ocaml_shout_set_password(value block, value password)
{
  CAMLparam2(block, password);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_password(s, String_val(password));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_password(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *password = shout_get_password(s);
  CAMLreturn(copy_string(password));
}

CAMLprim value ocaml_shout_set_mount(value block, value mount)
{
  CAMLparam2(block, mount);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_mount(s, String_val(mount));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_mount(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *mount = shout_get_mount(s);
  CAMLreturn(copy_string(mount));
}

CAMLprim value ocaml_shout_set_name(value block, value name)
{
  CAMLparam2(block, name);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_name(s, String_val(name));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_name(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *name = shout_get_name(s);
  CAMLreturn(copy_string(name));
}

CAMLprim value ocaml_shout_set_url(value block, value url)
{
  CAMLparam2(block, url);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_url(s, String_val(url));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_url(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *url = shout_get_url(s);
  CAMLreturn(copy_string(url));
}

CAMLprim value ocaml_shout_set_genre(value block, value genre)
{
  CAMLparam2(block, genre);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_genre(s, String_val(genre));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_genre(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *genre = shout_get_genre(s);
  CAMLreturn(copy_string(genre));
}

CAMLprim value ocaml_shout_set_user(value block, value username)
{
  CAMLparam2(block, username);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_user(s, String_val(username));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_user(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *username = shout_get_user(s);
  CAMLreturn(copy_string(username));
}

CAMLprim value ocaml_shout_set_agent(value block, value username)
{
  CAMLparam2(block, username);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_agent(s, String_val(username));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_agent(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *username = shout_get_agent(s);
  CAMLreturn(copy_string(username));
}

CAMLprim value ocaml_shout_set_description(value block, value desription)
{
  CAMLparam2(block, desription);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_description(s, String_val(desription));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_description(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *desription = shout_get_description(s);
  CAMLreturn(copy_string(desription));
}

CAMLprim value ocaml_shout_set_dumpfile(value block, value dumpfile)
{
  CAMLparam2(block, dumpfile);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_dumpfile(s, String_val(dumpfile));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_dumpfile(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  const char *dumpfile = shout_get_dumpfile(s);
  CAMLreturn(copy_string(dumpfile));
}

CAMLprim value ocaml_shout_set_audio_info(value block, value name, value val)
{
  CAMLparam3(block, name, val);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_audio_info(s, String_val(name), String_val(val));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_audio_info(value block, value name)
{
  CAMLparam2(block, name);
  shout_t *s = shout_of_block(block);
  const char *val = shout_get_audio_info(s, String_val(name));
  CAMLreturn(copy_string(val));
}

CAMLprim value ocaml_shout_set_public(value block, value public)
{
  CAMLparam2(block, public);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_public(s, (unsigned int)Int_val(public));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_public(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  unsigned int public = shout_get_public(s);
  CAMLreturn(public?Val_true:Val_false);
}

static int data_format_table[] = {SHOUT_FORMAT_VORBIS, SHOUT_FORMAT_MP3};

CAMLprim value ocaml_shout_set_format(value block, value data_format)
{
  CAMLparam2(block, data_format);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_format(s, data_format_table[Int_val(data_format)]);
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_format(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int data_format = shout_get_format(s);
  /* Not very clean: we should use data_format_table */
  CAMLreturn(Val_int(data_format));
}

static int protocol_table[] = {SHOUT_PROTOCOL_HTTP, SHOUT_PROTOCOL_XAUDIOCAST, SHOUT_PROTOCOL_ICY};

CAMLprim value ocaml_shout_set_protocol(value block, value protocol)
{
  CAMLparam2(block, protocol);
  shout_t *s = shout_of_block(block);
  int ret = shout_set_protocol(s, protocol_table[Int_val(protocol)]);
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_get_protocol(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int protocol = shout_get_protocol(s);
  /* Not very clean: we should use protocol_table */
  CAMLreturn(Val_int(protocol));
}

CAMLprim value ocaml_shout_open(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int ret = shout_open(s);
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_close(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int ret = shout_close(s);
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_send(value block, value data)
{
  CAMLparam2(block, data);
  shout_t *s = shout_of_block(block);
  int ret = shout_send(s, String_val(data), string_length(data));
  CAMLreturn(unit_or_error(ret));
}

CAMLprim value ocaml_shout_send_raw(value block, value data)
{
  CAMLparam2(block, data);
  shout_t *s = shout_of_block(block);
  int ret = shout_send_raw(s, String_val(data), string_length(data));
  CAMLreturn(Val_int(ret));
}

CAMLprim value ocaml_shout_sync(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  shout_sync(s);
  CAMLreturn(Val_unit);
}

CAMLprim value ocaml_shout_delay(value block)
{
  CAMLparam1(block);
  shout_t *s = shout_of_block(block);
  int delay = shout_delay(s);
  CAMLreturn(Val_int(delay));
}

CAMLprim value ocaml_shout_set_metadata(value block, value data)
{
  CAMLparam2(block, data);
  shout_t *s = shout_of_block(block);
  shout_metadata_t *metadata = shout_metadata_new();
  int i,ret;
  char *name, *val;
  
  for (i=0; i<Wosize_val(data); i++)
    {
      CAMLlocal1(c);
      c = Field(data, i);
      name = String_val(Field(c, 0));
      val = String_val(Field(c, 1));
      shout_metadata_add(metadata, name, val);
    }
  ret = shout_set_metadata(s, metadata);
  shout_metadata_free(metadata);
  CAMLreturn(unit_or_error(ret));
}
